#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

from setroubleshoot.signature import *
from setroubleshoot.util import *
from setroubleshoot.log import *
from avc import *

from string import Template
import sys
import os.path

class Plugin:
    """
    Each plugin object recognizes one or more access denials and
    presents a description of the denial to the user. Optionally,
    the plugin can provide a suggestion for allowing the access
    to the user.

    There are four user visible strings that arepart of each Plugin
    subclass (some or all of these can be changed by the plugin author):
     * summary: summary of the denial
     * problem_description: detailed description of the denial
     * fix_description: description of how to allow the denied access
     * fix_cmd: command that can be used to allow the access

    All of the strings will have a standard set of substitutions performed.
    Each keyword (proceeded by a '$' will be replace by a string) - see
    http://docs.python.org/lib/node109.html for more information. The
    keywords are:
     * $SOURCE_TYPE - type for the source of the avc (usually the
       process performing the operation).
     * $TARGET_TYPE - type for the target of the avc (the type of
       the object).
     * $SOURCE_PATH - source of the executable (from the exe or comm
       field of the exe). A full path is not always available.
     * $TARGET_PATH - path for the target object. A full path is not
       always available.
     * $TARGET_DIR - path of the containing directory for TARGET_PATH.
       Essentially os.path.dirname($TARGET_PATH)
     * $TARGET_CLASS - the object class for the target.
     * $PERMS - the permissions denied.
     * $SOURCE_PACKAGE - name of the package which contains the
       executable (from $SOURCE_PATH).
     * $PORTNUMBER - the port number for the connection denied.
    Additional subtitutions can be added with set_sub.

    You can also optional pass the name for a single boolean which will be
    used to set the $BOOLEAN subtitution into Plugin.__init__.
        """
    summary = ""
    problem_description = ""
    fix_description = ""
    fix_cmd = ""
        
    def __init__(self, name, category = None, bool_name = None):
        PROGNAME="setroubleshoot"

        import gettext
        gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
        gettext.textdomain(PROGNAME)
        try:
            gettext.install(PROGNAME, localedir="/usr/share/locale", unicode=1)
        except IOError:
            import __builtin__
            __builtin__.__dict__['_'] = unicode

        self.analysis_id = name
        self.category = category
        self.sub = {}
        self.set_sub("BOOLEAN", bool_name)
        self.priority=50
        self.is_bug=False
        self.src_rpms=[]
        self.tgt_rpms=[]
        
    def set_boolean(self, bool_name):
        self.set_sub("BOOLEAN", bool_name)

    def set_sub(self, name, value):
        if value:
            self.sub[name] = value

    def set_alt_path(self, path):
        if self.path is None:
            self.path = path

    def __process_avc(self):
        # Do basic processing of the avc to extract information that
        # most plugins will use. Only information that requires
        # processing or is part of the signature should be handled here
        # (e.g., determing the exe name). Simple information that can
        # unambiogously be retrieved from the original avc is not usually
        # processed (e.g., source type).
        #
        # The current fields that are processed and stored are.
        # * binary (either exe or comm)
        # * path (name or path)
        # * package

        self.path = self.avc.getPath()
        self.binary = self.avc.getBinary()
        self.src_rpms=[]
        self.tgt_rpms=[]

        if self.binary != None:
            self.source_pkg = get_rpm_nvr_by_file_path(self.binary)
            if self.source_pkg != '':
                self.src_rpms.append(self.source_pkg)
        else:
            self.source_pkg = ""
            
        if self.path != None:
            rpm = get_rpm_nvr_by_file_path(self.path)
            if rpm != '':
                self.tgt_rpms.append(rpm)

    def __set_standard_sub(self):
        # See Plugin.report for the list of standard subtitions
        self.sub["SOURCE_TYPE"] = self.avc.scontext.type
        self.sub["TARGET_TYPE"] = self.avc.tcontext.type
        self.sub["SOURCE_PATH"] = self.binary
        self.sub["TARGET_PATH"] = self.path

        if self.path is None:
            self.sub["TARGET_DIR"] = None
        else:
            self.sub["TARGET_DIR"] = os.path.dirname(self.path)

        self.sub["TARGET_CLASS"] = self.avc.tclass

        if self.avc.access is None:
            self.sub["ACCESS"] = None
        else:
            self.sub["ACCESS"] = ' '.join(self.avc.access)

        self.sub["SOURCE_PACKAGE"] = self.source_pkg
        self.sub["PORTNUMBER"] = self.avc.dest

        # Now validate, replace any None values with friendly string
        for key, value in self.sub.items():
            if value is None:
                self.sub[key] = _('<Unknown>')


    def prepare(self, avc):
        # Plugin.prepare is called immediately before Plugin.analyze to perform
        # avc processing that is handled for the plugin author.
        self.avc = avc
        self.rpms = []
        self.__process_avc()
        self.__set_standard_sub()

    def __process_results(self):
        self.__summary = Template(self.summary).safe_substitute(self.sub)
        self.__problem_description = Template(self.problem_description).safe_substitute(self.sub)
        self.__fix_description = Template(self.fix_description).safe_substitute(self.sub)
        self.__fix_cmd = Template(self.fix_cmd).safe_substitute(self.sub)

    def report(self, report_receiver):
        """
        Report a denial and solution to the fault server.
        """
        self.__process_results()
        environment = self.get_environment()

        solution = SEFaultSolution(
            summary=self.__summary,
            problem_description=self.__problem_description,
            fix_description=self.__fix_description,
            fix_cmd=self.__fix_cmd
            )

        siginfo = SEFaultSignatureInfo(
            sig=self.get_signature(),
            solution=solution,
            category = self.category,
            object_path=self.path,
            avc_list=[self.avc],
            is_bug=self.is_bug,
            plugin_name=self.analysis_id,
            src_rpm_list=self.src_rpms,
            tgt_rpm_list=self.tgt_rpms,
            environment=environment
            )

        report_receiver.report_problem(siginfo)
        
    def get_environment(self):
        environment = SEEnvironment()
        return environment
    
    def get_signature(self):
        avc = AVC(scontext=self.avc.scontext,
                  tcontext=self.avc.tcontext,
                  access=self.avc.access)

        sig = SEFaultSignature(
            analysis_id=self.analysis_id,
            avc_list=[avc],
            object_path=self.path)
        return sig

    def analyze(self):
        return False

    def get_summary(self):
        self.__process_results()
        return self.__summary

    def get_problem_description(self):
        self.__process_results()
        return self.__problem_description
        
    def get_fix_description(self):
        self.__process_results()
        return self.__fix_description
        
    def get_fix_cmd(self):
        self.__process_results()
        return self.__fix_cmd
        
    def set_priority(self, priority):
        self.priority = priority
        
    def get_priority(self):
        return self.priority 
        
        
