#
# Copyright (C) 2006 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['RunFaultServer',
           'get_host_database',
           'send_alert_notification',
           'send_update_notification',
          ]

PROGNAME="setroubleshoot"
import gettext
gettext.bindtextdomain(PROGNAME, "/usr/share/locale")
gettext.textdomain(PROGNAME)
try:
    gettext.install(PROGNAME, "/usr/share/locale", 1)
except IOError:
    import __builtin__
    __builtin__.__dict__['_'] = unicode

from setroubleshoot.access_control import *
from setroubleshoot.avc_audit import *
from setroubleshoot.config import cfg
from setroubleshoot.errcode import *
from setroubleshoot.email_alert import *
from setroubleshoot.log import *
from setroubleshoot.rpc import *
from setroubleshoot.rpc_interfaces import *
from setroubleshoot.signature import *
from setroubleshoot.util import *

import errno
import gobject
import os
import pwd
import Queue
import signal
import socket
from stat import *
import sys
import syslog
import threading
from types import *


#------------------------------ Variables -------------------------------

host_database = None
analysis_queue = None

#------------------------------ Utility Functions -----------------------------

def sighandler(signum, frame):
    log_server.debug("received signal=%s", signum)
    import setroubleshoot.config as config
    if signum == signal.SIGHUP:
        log_server.warning("reloading configuration file")
        config.config_init()
        return
    
    print "Exiting . . . "
    sys.exit()

def get_host_database():
    return host_database

def send_alert_notification(siginfo):
    for client in connection_pool.clients('sealert'):
        client.alert(siginfo)

# FIXME: this needs to be database specific
def send_update_notification(criteria):
    for client in connection_pool.clients('sealert'):
        client.signatures_updated(criteria)

#------------------------------------------------------------------------------

class SETroubleshootDatabase:
    def __init__(self, filename):
        self.database_dir = cfg.get('database','database_dir')
        self.filename = filename
        self.filepath = os.path.join(self.database_dir, self.filename)
        self.lock = threading.Lock()
        self.load()

    def load(self):
        self.sigs = SEFaultSignatureSet()
        
        if os.path.exists(self.filepath):
           stat_info = os.stat(self.filepath)
           if stat_info[ST_SIZE] > 0:
               self.sigs.read_xml_file(self.filepath, 'sigs')

    def save(self):
        log_database.debug("writing database (%s)", self.filepath)
        self.sigs.write_xml('sigs', self.filepath)
        
    def acquire(self):
        self.lock.acquire()

    def release(self):
        self.lock.release()
    
    def lookup_signature(self, sig):
        siginfo = None

        matches = self.sigs.match_signatures(sig)
        log_sig.debug("lookup_signature: found %d matches with scores %s",
                      len(matches), ",".join(["%.2f" % x.score for x in matches]))
        if len(matches) == 0:
            raise ProgramError(ERR_NO_SIGNATURE_MATCH)
        elif len(matches) > 1:
            raise ProgramError(ERR_MULTIPLE_SIGNATURE_MATCH)
        siginfo = matches[0].siginfo
        return siginfo

    def add_siginfo(self, siginfo):
        return self.sigs.add_siginfo(siginfo)

    def remove_siginfo(self, siginfo):
        self.sigs.remove_siginfo(siginfo)

    def get_xml_nodes_filtered(self, criteria):
        root = self.sigs.get_xml_nodes('sigs')
        return root

#-----------------------------------------------------------------------------

class ConnectionPool(object):
    def __init__(self):
        self.client_pool = {}

    def add_client(self, handler):
        if self.client_pool.has_key(handler):
            log_rpc.warning("add_client: client (%s) already in client pool" % handler)
            return
        self.client_pool[handler] = None

    def remove_client(self, handler):
        if not self.client_pool.has_key(handler):
            log_rpc.warning("remove_client: client (%s) not in client pool" % handler)
            return
        del(self.client_pool[handler])

    def clients(self, channel_type=None):
        for client in self.client_pool:
            if channel_type is None:
                yield client
            elif client.channel_type == channel_type:
                yield client

    def close_all(self, channel_type=None):
        for client in self.client_pool(channel_type):
            client.close_connection()

connection_pool = ConnectionPool()

#------------------------------------------------------------------------------

class Analyze(threading.Thread):
    def __init__(self,queue):
        # parent class constructor
        threading.Thread.__init__(self)
        self.queue=queue
        self.plugins = load_plugins()
        log_avc.debug("Number of Plugins = %d", len(self.plugins))
        
    def run(self):
        while True:
            avc, report_receiver = self.queue.get()
            for plugin in self.plugins:
                try:	
                    plugin.prepare(avc)
                    if plugin.analyze():
                        plugin.report(report_receiver)
                        break;
                    
                except Exception, e:
                    log_avc.exception("Plugin Exception %s " % plugin.analysis_id) 
                    self.plugins.remove(plugin)

    
#------------------------------------------------------------------------------

class PluginReportReceiver(object):
    def __init__(self, database):
        self.database = database

    def report_problem(self, siginfo):
        timestamp = TimeStamp()
        try:
            database_siginfo = self.database.lookup_signature(siginfo.sig)
            log_database.debug("signature found in database")
        except ProgramError, e:
            if e.errno == ERR_NO_SIGNATURE_MATCH:
                log_database.debug("not in database yet")
                siginfo.local_id=self.database.sigs.generate_local_id()
                siginfo.first_seen_date = str(timestamp)
                database_siginfo = self.database.add_siginfo(siginfo)
            else:
                raise
        # FIXME: merge new siginfo data with what is already in the database
        siginfo = database_siginfo

        siginfo.last_seen_date = str(timestamp)
        siginfo.report_count += 1
        self.database.save()
        return siginfo

#------------------------------------------------------------------------------

class AlertPluginReportReceiver(PluginReportReceiver):
    def __init__(self, database):
        super(AlertPluginReportReceiver, self).__init__(database)

    def report_problem(self, siginfo):
        siginfo = super(AlertPluginReportReceiver, self).report_problem(siginfo)
        recipients = cfg.get('email','recipients')
        if len(recipients) != 0:
            to_addrs = [addr.strip() for addr in recipients.split(',')]
            for addr in to_addrs:
                action = siginfo.evaluate_filter_for_user(addr)
                if action == "ignore":
                    to_addrs.remove(addr)
                else:
                    siginfo.update_user_filter(addr, FILTER_ALWAYS)
                    self.database.save()

            if len(to_addrs):
                email_alert(siginfo, to_addrs)
        
        log_alert.debug("sending alert to all clients")
        send_alert_notification(siginfo)
        send_update_notification('')

        # FIXME: should this be using our logging objects in log.py?
        syslog.syslog(siginfo.solution.summary + \
                      _(" See audit.log for complete SELinux messages. id = %s" % siginfo.local_id ))
        return siginfo

#-----------------------------------------------------------------------------

class ClientConnectionHandler(RpcChannel, SETroubleshootServerInterface, SEAlertInterface):
    def __init__(self):
        RpcChannel.__init__(self, self.on_connection_state_change, 'sealert')
        self.database = get_host_database()
        self.access = ServerAccess()
        self.username = None
        self.uid = None
        self.gid = None

    def on_connection_state_change(self, connection_state):
        log_rpc.debug("on_connection_state_change: state=%s channel=%s type=%s",
                      connection_state, self.channel_name, self.channel_type)
        if connection_state == 'closed':
            connection_pool.remove_client(self)
        elif connection_state == 'open':
            connection_pool.add_client(self)
        elif connection_state == 'retry':
            pass
        elif connection_state == 'hup':
            pass
        elif connection_state == 'error':
            pass
        elif connection_state == 'invalid':
            pass
        elif connection_state == 'timeout':
            pass
        else:
            log_rpc.error("on_connection_state_change: unknown state (%s) channel=%s type=%s",
                          connection_state, self.channel_name, self.channel_type)

    def open(self, socket, socket_address):
        if self.connection_state == 'open':
            return True
        self.socket = socket
        self.socket_address = socket_address
        self.set_connection_state('open')
        self.io_watch_id = gobject.io_add_watch(self.socket, io_input_conditions, self.handle_client_io)


    # ---- Handlers ----

    def analyze_logfile(self, logfile_path, options):
        log_rpc.debug("analyze_logfile: file=%s options=%s",
                      logfile_path, options)
        
        logfile_basename = os.path.basename(logfile_path)
        database_filename = logfile_basename + '.xml'
        if os.path.exists(database_filename):
            os.remove(database_filename)

        logfile_database = SETroubleshootDatabase(database_filename)
        report_receiver = PluginReportReceiver(logfile_database)
        siginfo = parse_audit_logfile(logfile_path, report_receiver)

        return None
        
    def delete_signature(self, sig):
        log_rpc.debug("delete_signature: sig=%s", sig)

        siginfo = self.database.lookup_signature(sig)
        self.database.remove_siginfo(siginfo)
        self.database.save()
        send_update_notification('')

        return None
        
    def logon(self, type, username, password):
        log_rpc.debug("logon(%s) type=%s username=%s", self, type, username)
        self.channel_type = type
        self.channel_name = username
        self.user = self.database.sigs.users.get_user(username)
        if self.user is None:
            self.user = self.database.sigs.users.add_user(username)
            self.database.save()

        return None

    def evaluate_alert_filter(self, sig, username):
        log_rpc.debug("evaluate_alert_filter: username=%s sig=%s",
                      username, sig)

        siginfo = self.database.lookup_signature(sig)
        action = siginfo.evaluate_filter_for_user(username)
        return [action]

    def query_alerts(self, criteria):
        log_rpc.debug("query_alerts: criteria=%s", criteria)

        siginfo = host_database.get_xml_nodes_filtered('*')
        return [siginfo]

    def set_user_data(self, sig, username, item, data):
        log_rpc.debug("set_user_data: username=%s item=%s data=%s sig=\n%s",
                      username, item, data, sig)

        siginfo = self.database.lookup_signature(sig)
        user_data = siginfo.get_user_data(username)
        result_code, result_msg = user_data.update_item(item, data)
        if result_code == 0:
            self.database.save()
        else:
            raise ProgramError(result_code, result_msg)
        return None

    def set_filter(self, sig, username, filter_type, data = "" ):
        log_rpc.debug("set_filter: username=%s filter_type=%s sig=\n%s",
                      username, filter_type, sig)

        siginfo = self.database.lookup_signature(sig)
        siginfo.update_user_filter(username, filter_type, data)
        self.database.save()
        return None


#------------------------------------------------------------------------------

def RunFaultServer():
    gobject.threads_init()

    # FIXME
    global host_database, analysis_queue

    signal.signal(signal.SIGHUP, sighandler)
    signal.signal(signal.SIGQUIT, sighandler)
    signal.signal(signal.SIGTERM, sighandler)

    #interface_registry.dump_interfaces()

    try:
        # FIXME: should this be using our logging objects in log.py?
        # currently syslog is only used for putting an alert into
        # the syslog with it's id

        pkg_name = cfg.get('general','pkg_name')
        syslog.openlog(pkg_name)

        # Create a database local to this host

        host_database = SETroubleshootDatabase(cfg.get('database','filename'))

        # Attach the local database to an object which will send alerts
        # specific to this host

        alert_receiver = AlertPluginReportReceiver(host_database)

        # Create a synchronized queue for analysis requests

        analysis_queue = Queue.Queue(0)

        # Create a thread to peform analysis, it takes AVC objects off
        # the the analysis queue and runs the plugins against the
        # AVC. Analysis requests in the queue may arrive from a
        # variety of places; from the audit system, from a log file
        # scan, etc. The disposition of the analysis (e.g. where the
        # results of the analysis are to go) are included in the queued
        # object along with the data to analyze.

        analyze_thread = Analyze(analysis_queue)
        analyze_thread.setDaemon(True)
        analyze_thread.start()
    
        # Create a thread to receive messages from the audit system.
        # This is a time sensitive operation, the primary job of this
        # thread is to receive the audit message as quickly as
        # possible and return to listening on the audit socket. When
        # it receives a complete audit event it places it in the
        # analysis queue where another thread will process it
        # independently.

        audit_socket_thread = AuditSocketReceiver(analysis_queue, alert_receiver)
        audit_socket_thread.setDaemon(True)
        audit_socket_thread.start()

        # Create a server to listen for alert clients and then run.
        listening_server = ListeningServer(get_server_address('local_fault_server'), ClientConnectionHandler)
        listening_server.open()
        main_loop = gobject.MainLoop()
        main_loop.run()

    except KeyboardInterrupt, e:
        pass

    except SystemExit, e:
        pass

    except Exception, e:
        log_rpc.exception("exception %s: %s",e.__class__.__name__, str(e))

if __name__=='__main__':
    RunFaultServer()

